% Exercise 1
% -------------------------
clear all, close all
%Process initialization
A_tank=0.1;
Kp=0.002;
h_init=0.5;
h=0.5;
F_out=0.005;

% time-step
Ts=0.1;
t_start=0;
t_end=100;
t_array=[t_start:Ts:(t_end-Ts)];
N=length(t_array);

% model in sybolic for jacobian calculation
syms x_1 x_2 u f1 f2 g;
f1=x_1+Ts/A_tank*(Kp*u-x_2);
f2=x_2;
g=x_1;
x=[x_1;x_2];
f=[f1;f2];
A=jacobian(f,x);
C=jacobian(g,x);

% PID controller
u_0=2.5;
u_max=5;
u_min=0;
Kc=10;
Ti=10;
Td=0;
h_sp=0.5;
Esum=0;

% states
x1_init=h; x2_init=F_out; y_init=x1_init;%define standard names
x1=x1_init; x2=x2_init;y=y_init;
x1_array=zeros(1,N);x2_array=zeros(1,N); % storage of real values
x1_p=zeros(1,N);x2_p=zeros(1,N);y_p=zeros(1,N); % storage of predicted state estimate
x1_c=zeros(1,N);x2_c=zeros(1,N); % storage of corrected state
x_p=[x1_p;x2_p];x_c=[x1_c;x2_c]; % vectorization
x_init=[x1_init;x2_init];
x_p_k=x_init;
x_c_k=x_init;

% for Kalman filter 
Q1=0.01;Q2=0.000001;
Q=[Q1 0;0 Q2]; % assume constant variance of state variables
R=[0.0001]; % assume constant variance of the mea
G=eye(2);


P_p_init = diag(x_init.*x_init);
P_p_k = P_p_init;


% -----------------------
for k=1:N
    t_k=k*Ts;
    w1=normrnd(0,sqrt(Q1));
    w2=normrnd(0,sqrt(Q2));
    x1=x1+w1;x2=x2+w2;
    if x1<=0;
        x1=0;
    end
    if x1>=1;
        x1=1;
    end
     if x2<=0;
        x2=0;
    end
    if x2>=0.01;
        x2=0.01;
    end
    v=normrnd(0,sqrt(R));    %random measurement noise
    y_m=x1+v;                
   
    e_y_sp=h_sp-y_m;          % error from the setpoint
    Esum=Esum+e_y_sp;
    u=u_0+Kc*e_y_sp+Kc/Ti*Esum;   % controller update
    if u>=5;
        u=5;
    end
    if u<=0;
        u=0;
    end
    x1_array(k)=x1; % store the real values for plot
    x2_array(k)=x2;
    if x2_p(k)<=0;
        x2_p(k)=0;
    end
    if x2_p(k)>=0.01;
        x2_p(k)=0.01;
    end
    % Kalman gain K=[K11;K21]
    K_k = P_p_k*C'*inv(C*P_p_k*C'+R);
    
    
    % Update of states based on process model and Kalman filter
    x1_p(1)=h_init;x2_p(1)=F_out; % initial value
    y_p(k)=x1_p(k);
    e(k)=y_m-y_p(k);
    x1_c(k)=x1_p(k)+K_k(1)*e(k);
    x2_c(k)=x2_p(k)+K_k(2)*e(k);
    if x2_c(k)<=0;
        x2_c(k)=0;
    end
    if x2_c(k)>=0.01;
        x2_c(k)=0.01;
    end
    if k<=N-1
        x1_p(k+1)=x1_c(k)+Ts/A_tank*(Kp*u-x2_c(k));
        x2_p(k+1)=x2_c(k);
    end
 
 
    % auto-covaraince of error for corrected estimate:
    P_c_k=(eye(2)-K_k*C)*P_p_k;
    
    % Update of the auto-covariance of error for predicted estimate 
    P_p_kp1=A*P_c_k*A'+Q;
    
    % Time shift
    P_p_k=P_p_kp1;
    
    
end
% plot x1
figure 
ax1=subplot(2,1,1)
plot(ax1,t_array,x1_array,t_array,x1_p, 'r--',t_array,x1_c,'g:');
xlabel('Time');
ylabel('Level (m)');
legend('real','predicted','corrected');
title ('Kalman Filter estimation for x1-Level');
% plot x2
ax2=subplot(2,1,2)
plot(ax2,t_array,x2_array,t_array,x2_p, 'r--',t_array,x2_c,'g:');
xlabel('Time');
ylabel('Outflow (m3/s)');
legend('real','predicted','corrected');
title ('Kalman Filter estimation for x2-Outflow');
